import asyncio
import copy
from dataclasses import dataclass, field
from functools import partial
import logging
from typing import Callable, Optional
from weakref import WeakValueDictionary

from ..base_manager import BaseManager
from ..exceptions import InvalidSessionError
from ..network.connection import ConnectionState, PeerConnection, ServerConnection
from ..events import (
    build_message_map,
    on_message,
    AdminMessageEvent,
    BlockListChangedEvent,
    ConnectionStateChangedEvent,
    Event,
    EventBus,
    FriendListChangedEvent,
    KickedEvent,
    MessageReceivedEvent,
    PrivateMessageEvent,
    PrivilegedUsersEvent,
    PrivilegedUserAddedEvent,
    PrivilegesUpdateEvent,
    UserInfoUpdateEvent,
    UserStatsUpdateEvent,
    UserStatusUpdateEvent,
    UserTrackingStateChangedEvent,
    UserTrackingEvent,
    UserUntrackingEvent,
    SessionInitializedEvent,
    SessionDestroyedEvent,
)
from ..protocol.messages import (
    AddPrivilegedUser,
    AddUser,
    AdminMessage,
    PrivateChatMessage,
    PrivateChatMessageAck,
    CheckPrivileges,
    GetUserStatus,
    GetUserStats,
    Kicked,
    PeerSearchReply,
    PeerUserInfoReply,
    PrivilegedUsers,
    RemoveUser,
    SetStatus,
)
from .model import (
    BlockingFlag,
    ChatMessage,
    UploadPermissions,
    User,
    UserStatus,
    TrackingFlag,
    TrackingState,
)
from ..network.network import Network
from ..settings import Settings
from ..session import Session
from ..tasks import BackgroundTask
from ..utils import cancel_task


logger = logging.getLogger(__name__)


RETRY_TIMEOUT_NET_ERROR = 10
RETRY_TIMEOUT_NON_EXISTING_USER = 600


@dataclass(slots=True)
class TrackingRequest:
    operation: Callable[[TrackingFlag], None]
    flag: TrackingFlag
    handled: asyncio.Event = field(default_factory=asyncio.Event)


@dataclass(slots=True)
class TrackedUser:
    user: User
    flags: TrackingFlag = TrackingFlag(0)
    state: TrackingState = TrackingState.UNTRACKED

    queue: asyncio.Queue[TrackingRequest] = field(default_factory=asyncio.Queue)
    task: Optional[asyncio.Task] = None
    retry_task: Optional[asyncio.Task] = None

    def add_flag(self, flag: TrackingFlag):
        self.flags |= flag

    def remove_flag(self, flag: TrackingFlag):
        self.flags &= ~flag


@dataclass(slots=True)
class UserManagementContext:
    friends: set[str]
    blocked: dict[str, BlockingFlag]


class UserManager(BaseManager):
    """Class responsible for handling user messages and storing users"""

    def __init__(self, settings: Settings, event_bus: EventBus, network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._tracking_manager: UserTrackingManager = UserTrackingManager(
            settings, event_bus, network)

        self._session: Optional[Session] = None

        self._MESSAGE_MAP = build_message_map(self)

        self._users: WeakValueDictionary[str, User] = WeakValueDictionary()
        self._privileged_users: set[str] = set()

        self._management_task: BackgroundTask = BackgroundTask(
            interval=1.0,
            task_coro=self._management_job,
            name='user-management-task',
            context=UserManagementContext(
                friends=self._settings.users.friends.copy(),
                blocked=self._settings.users.blocked.copy()
            )
        )

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)
        self._event_bus.register(
            FriendListChangedEvent, self._on_friend_list_changed)

    @property
    def users(self) -> dict[str, User]:
        return dict(self._users)

    @property
    def privileged_users(self) -> set[str]:
        return self._privileged_users

    async def start(self):
        self._management_task.start()

    async def stop(self) -> list[asyncio.Task]:
        cancel_tasks = []
        if task := self._management_task.cancel():
            cancel_tasks.append(task)

        tracking_tasks = self._tracking_manager.stop()
        cancel_tasks.extend(tracking_tasks)

        return cancel_tasks

    def get_self(self) -> User:
        """Returns the user object for the current session"""
        if not self._session:  # pragma: no cover
            raise InvalidSessionError("user is not logged in")

        return self.get_user_object(self._session.user.name)

    def get_user_object(self, username: str) -> User:
        """Gets a :class:`.User` object for given ``username``, if the user is
        not stored it will be created and stored.

        A user will be stored only if there is a strong reference to the
        :class:`.User` object, otherwise it will be removed

        :param username: Name of the user
        :return: a :class:`.User` object
        """
        if username not in self._users:
            # Don't simplify this by assigning it directly to the dict; a strong
            # reference needs to be kept otherwise the object would be
            # immediatly removed
            user = User(
                name=username,
                privileged=username in self._privileged_users
            )
            self._users[username] = user
            return user

        return self._users[username]

    def is_tracked(self, username: str) -> bool:
        """Returns whether a user is currently being tracked"""
        return self._tracking_manager.get_tracking_state(username) == TrackingState.TRACKED

    def reset_users(self):
        """Performs a reset on all users. This method is called when the
        connection with the server is disconnected and shouldn't be called
        directly
        """
        self._users = WeakValueDictionary()
        self._privileged_users = set()

    async def track_user(self, username: str, flag: TrackingFlag = TrackingFlag.REQUESTED):
        """Requests to start tracking a user. Tracking of a user will be handled
        in the background

        :param username: user to track
        :param flag: tracking flag to add from the user
        """
        user = self.get_user_object(username)
        self._tracking_manager.track_user(user, flag=flag)

    async def untrack_user(self, username: str, flag: TrackingFlag = TrackingFlag.REQUESTED):
        """Requests to stop tracking a user. Tracking of a user will be handled
        in the background.

        Untracking does not necesserily mean that the library stops receiving
        user updates, the library could internally still be wanting to track
        a user for example for managing transfers

        :param username: user to track
        :param flag: tracking flag to add from the user
        """
        user = self.get_user_object(username)
        self._tracking_manager.untrack_user(user, flag=flag)

    async def track_friend(self, username: str):
        """Request to track a friend with given username"""
        await self.track_user(username, TrackingFlag.FRIEND)

    async def track_friends(self):
        """Starts tracking the users defined in the friends list"""
        tasks = [
            self.track_friend(friend)
            for friend in self._settings.users.friends
        ]

        await asyncio.gather(*tasks, return_exceptions=True)

    async def untrack_friend(self, username: str):
        """Request to stop tracking a friend with given username"""
        await self.untrack_user(username, TrackingFlag.FRIEND)

    def get_tracking_flags(self, username: str) -> TrackingFlag:
        """Returns current desired tracking flags of the user

        :param username: username of the user
        """
        return self._tracking_manager.get_tracking_flags(username)

    def get_tracking_state(self, username: str) -> TrackingState:
        """Returns current tracking state of the user

        :param username: username of the user
        """
        return self._tracking_manager.get_tracking_state(username)

    async def _management_job(self, context: UserManagementContext):
        events: list[Event] = []

        if context.friends != self._settings.users.friends:
            events.append(
                FriendListChangedEvent(
                    added=self._settings.users.friends - context.friends,
                    removed=context.friends - self._settings.users.friends
                )
            )

        if context.blocked != self._settings.users.blocked:
            changes = {}

            # Unblocked through removal
            for username in context.blocked.keys() - self._settings.users.blocked.keys():
                changes[username] = (context.blocked[username], BlockingFlag.NONE)

            # Changed flag or added
            for username, flags in self._settings.users.blocked.items():
                old_flags = context.blocked.get(username, BlockingFlag.NONE)
                if flags != old_flags:
                    changes[username] = (old_flags, flags)

            events.append(BlockListChangedEvent(changes=changes))

        if events:
            context.friends = self._settings.users.friends.copy()
            context.blocked = self._settings.users.blocked.copy()

        for event in events:
            await self._event_bus.emit(event)

    @on_message(AdminMessage.Response)
    async def _on_admin_message(self, message: AdminMessage.Response, connection: ServerConnection):
        await self._event_bus.emit(
            AdminMessageEvent(
                message.message,
                raw_message=message
            )
        )

    @on_message(Kicked.Response)
    async def _on_kicked(self, message: Kicked.Response, connection: ServerConnection):
        await self._event_bus.emit(KickedEvent(raw_message=message))

    @on_message(PrivateChatMessage.Response)
    async def _on_private_message(self, message: PrivateChatMessage.Response, connection: ServerConnection):

        await self._network.send_server_messages(
            PrivateChatMessageAck.Request(message.chat_id)
        )

        # For blocked users, ack the message but don't emit an event
        if self._settings.users.is_blocked(message.username, BlockingFlag.PRIVATE_MESSAGES):
            return

        user = self.get_user_object(message.username)
        chat_message = ChatMessage(
            id=message.chat_id,
            timestamp=message.timestamp,
            user=user,
            message=message.message,
            is_direct=bool(message.is_direct)
        )

        await self._event_bus.emit(
            PrivateMessageEvent(
                message=chat_message,
                raw_message=message
            )
        )

    # State related messages
    @on_message(CheckPrivileges.Response)
    async def _on_check_privileges(self, message: CheckPrivileges.Response, connection: ServerConnection):
        if not self._session:  # pragma: no cover
            raise InvalidSessionError("user is not logged in")

        self.privileges_time_left = message.time_left
        self._session.privileges_time_left = message.time_left

        await self._event_bus.emit(
            PrivilegesUpdateEvent(
                time_left=message.time_left,
                raw_message=message
            )
        )

    @on_message(PrivilegedUsers.Response)
    async def _on_privileged_users(self, message: PrivilegedUsers.Response, connection: ServerConnection):

        for user in self._users.values():
            user.privileged = user.name in message.users

        self._privileged_users = set(message.users)

        await self._event_bus.emit(
            PrivilegedUsersEvent(
                users=list(map(self.get_user_object, message.users)),
                raw_message=message
            )
        )

    @on_message(AddPrivilegedUser.Response)
    async def _on_add_privileged_user(self, message: AddPrivilegedUser.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)
        user.privileged = True

        await self._event_bus.emit(PrivilegedUserAddedEvent(user))

    @on_message(AddUser.Response)
    async def _on_add_user(self, message: AddUser.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)
        if message.exists:
            user.status = UserStatus(message.status)
            if message.user_stats:
                user.update_from_user_stats(message.user_stats)
            user.country = message.country_code

    @on_message(GetUserStatus.Response)
    async def _on_get_user_status(self, message: GetUserStatus.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)

        before = copy.deepcopy(user)

        user.status = UserStatus(message.status)
        user.privileged = message.privileged

        await self._event_bus.emit(
            UserStatusUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(GetUserStats.Response)
    async def _on_get_user_stats(self, message: GetUserStats.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)

        before = copy.deepcopy(user)

        user.update_from_user_stats(message.user_stats)

        await self._event_bus.emit(
            UserStatsUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(PeerUserInfoReply.Request)
    async def _on_peer_user_info_reply(self, message: PeerUserInfoReply.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerUserInfoReply for a connection that wasn't properly initialized")
            return

        user = self.get_user_object(connection.username)

        before = copy.deepcopy(user)

        user.description = message.description
        user.picture = message.picture
        user.upload_slots = message.upload_slots
        user.queue_length = message.queue_size
        user.has_slots_free = message.has_slots_free
        if message.upload_permissions is None:
            user.upload_permissions = UploadPermissions.UNKNOWN
        else:
            user.upload_permissions = UploadPermissions(message.upload_permissions)

        await self._event_bus.emit(
            UserInfoUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(PeerSearchReply.Request)
    async def _on_peer_search_reply(self, message: PeerSearchReply.Request, connection: PeerConnection):
        user = self.get_user_object(message.username)
        user.avg_speed = message.avg_speed
        user.queue_length = message.queue_size
        user.has_slots_free = message.has_slots_free

    # Listeners

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSED:
            self.reset_users()

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session
        await self._network.send_server_messages(
            CheckPrivileges.Request(),
            SetStatus.Request(UserStatus.ONLINE.value)
        )
        # Due to a bug in the protocol a GetUserStatus message for ourself is
        # never returned and it needs to be set manually
        self.get_self().status = UserStatus.ONLINE

        # Tracking does not actually work for the self user from a server point
        # of view but is kept here for convenience
        await self.track_user(self._session.user.name, TrackingFlag.FRIEND)

        # Perform AddUser for all in the friendlist
        await self.track_friends()

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None

    async def _on_friend_list_changed(self, event: FriendListChangedEvent):
        if not self._session:  # pragma: no cover
            return

        tasks = []
        for added_friend in event.added:
            tasks.append(self.track_friend(added_friend))

        for removed_friend in event.removed:
            tasks.append(self.untrack_friend(removed_friend))

        await asyncio.gather(*tasks, return_exceptions=True)


class UserTrackingManager:
    """Keeps the user tracking state in sync with the server"""

    def __init__(self, settings: Settings, event_bus: EventBus, network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network

        self._tracked_users: dict[str, TrackedUser] = dict()

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)

    def get_tracking_flags(self, username: str) -> TrackingFlag:
        """Returns current desired tracking flags of the user with given
        username

        :param username: username of the user
        """
        if username in self._tracked_users:
            return self._tracked_users[username].flags
        return TrackingFlag(0)

    def get_tracking_state(self, username: str) -> TrackingState:
        """Returns the current tracking state of the given user

        """
        if username in self._tracked_users:
            return self._tracked_users[username].state

        return TrackingState.UNTRACKED

    def track_user(self, user: User, flag: TrackingFlag = TrackingFlag.REQUESTED) -> TrackingRequest:
        tracked_user = self._get_tracked_user_object(user)

        request = TrackingRequest(tracked_user.add_flag, flag)
        tracked_user.queue.put_nowait(request)

        return request

    def untrack_user(self, user: User, flag: TrackingFlag = TrackingFlag.REQUESTED) -> Optional[TrackingRequest]:
        if user.name not in self._tracked_users:
            return None

        tracked_user = self._tracked_users[user.name]

        request = TrackingRequest(tracked_user.remove_flag, flag)
        tracked_user.queue.put_nowait(request)

        return request

    async def _tracking_task(self, tracked_user: TrackedUser):
        while True:
            request = await tracked_user.queue.get()

            previous_flags = tracked_user.flags
            request.operation(request.flag)
            is_retry = request.flag == TrackingFlag(0)

            if tracked_user.flags == TrackingFlag(0):
                # Ensure retry does not get scheduled again if we no longer
                # desire to track the user
                await cancel_task(tracked_user.retry_task)

                # Prevent RemoveUser from being called multiple times if there
                # are multiple entries on the queue
                if previous_flags != TrackingFlag(0):
                    await self._request_untracking(tracked_user)
                    await self._set_tracking_state(
                        tracked_user,
                        TrackingState.UNTRACKED
                    )

                # Only if the queue is empty can the user be removed from the
                # tracked users and task ended. It's possible there's a request
                # to start tracking again, the user cannot be removed otherwise
                # there is a "ghost" user being tracked
                if tracked_user.queue.empty():
                    request.handled.set()
                    return

            elif previous_flags == TrackingFlag(0) or is_retry:
                retry_timeout, retry_reason, response = await self._request_tracking(tracked_user)

                if retry_timeout:
                    logger.debug(
                        "failed to track user %s retrying in %d seconds. reason : %r",
                        tracked_user.user.name, retry_timeout, retry_reason
                    )

                    # Cancelling probably shouldn't be necessary but just doing
                    # it for safety
                    await cancel_task(tracked_user.retry_task)
                    tracked_user.retry_task = asyncio.create_task(
                        self._request_retry(tracked_user, retry_timeout))

                    await self._set_tracking_state(
                        tracked_user,
                        TrackingState.RETRY_PENDING,
                        message=response
                    )

                else:
                    await self._set_tracking_state(
                        tracked_user,
                        TrackingState.TRACKED,
                        message=response
                    )

            request.handled.set()

    async def _request_retry(self, tracked_user: TrackedUser, timeout: float):
        await asyncio.sleep(timeout)
        request = TrackingRequest(tracked_user.add_flag, TrackingFlag(0))
        tracked_user.queue.put_nowait(request)

    async def _request_tracking(
            self, tracked_user: TrackedUser) -> tuple[Optional[float], Optional[str], Optional[AddUser.Response]]:

        username = tracked_user.user.name
        try:
            await self._network.send_server_messages(AddUser.Request(username))

        except Exception:
            return RETRY_TIMEOUT_NET_ERROR, "failed to send tracking message", None

        try:
            response = await self._network.wait_for_server_message(
                AddUser.Response,
                fields={'username': username},
                timeout=10
            )

        except TimeoutError:
            return RETRY_TIMEOUT_NET_ERROR, "timeout waiting for tracking response", None

        except Exception as exc:
            return RETRY_TIMEOUT_NET_ERROR, f"unknown error waiting for tracking response : {exc!r}", None

        if not response.exists:
            return RETRY_TIMEOUT_NON_EXISTING_USER, "user does not exist", response

        return None, None, response

    async def _request_untracking(self, tracked_user: TrackedUser):
        username = tracked_user.user.name
        try:
            await self._network.send_server_messages(RemoveUser.Request(username))

        except Exception as exc:
            logger.debug(
                "failed to send untracking request for user %s : %r",
                username, exc
            )

    async def _set_tracking_state(
            self, tracked_user: TrackedUser, state: TrackingState,
            message: Optional[AddUser.Response] = None):

        tracked_user.state = state

        # Deprecated events
        if state == TrackingState.TRACKED:
            await self._event_bus.emit(
                UserTrackingEvent(tracked_user.user, message))  # type: ignore

        elif state == TrackingState.UNTRACKED:
            await self._event_bus.emit(
                UserUntrackingEvent(tracked_user.user))

        await self._event_bus.emit(
            UserTrackingStateChangedEvent(tracked_user.user, state, message))

    def _get_tracked_user_object(self, user: User) -> TrackedUser:
        """Gets or creates a tracked user object"""
        if user.name in self._tracked_users:
            tracked_user = self._tracked_users[user.name]

        else:
            tracked_user = TrackedUser(user)
            tracked_user.task = asyncio.create_task(
                self._tracking_task(tracked_user))
            tracked_user.task.add_done_callback(
                partial(self._on_tracking_task_done, tracked_user))

            self._tracked_users[user.name] = tracked_user

        return tracked_user

    def _on_tracking_task_done(self, tracked_user: TrackedUser, task: asyncio.Task):
        try:
            task.result()

        except asyncio.CancelledError:
            pass

        except Exception as exc:
            logger.warning(
                "unknown exception in tracking task for user %s : %r",
                tracked_user.user.name, exc
            )

        finally:
            self._tracked_users.pop(tracked_user.user.name, None)

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSED:
            tasks = self.stop()
            if tasks:
                await asyncio.gather(*self.stop(), return_exceptions=True)

    def stop(self) -> list[asyncio.Task]:
        tasks = []
        for tracked_user in self._tracked_users.values():
            if tracked_user.task:
                tracked_user.task.cancel()
                tasks.append(tracked_user.task)

            if tracked_user.retry_task:
                tracked_user.retry_task.cancel()
                tasks.append(tracked_user.retry_task)

        return tasks



from __future__ import annotations
from dataclasses import dataclass, field
import sys
from typing import Optional

from ..user.model import User


if sys.version_info >= (3, 11):
    slots_params = {'slots': True, 'weakref_slot': True}

else:
    slots_params = {}


@dataclass(**slots_params)
class Room:
    name: str
    private: bool = False
    users: list[User] = field(default_factory=list)
    """Current list of joined users"""
    joined: bool = False
    user_count: int = 0
    tickers: dict[str, str] = field(default_factory=dict)
    """Room tickers (room wall)"""

    # Only for private rooms
    members: set[str] = field(default_factory=set)
    """For private rooms, names of members of the room (excludes owner)"""
    owner: Optional[str] = None
    """For private rooms, name of the room owner"""
    operators: set[str] = field(default_factory=set)
    """For private rooms, names of operators"""

    def add_user(self, user: User):
        if user not in self.users:
            self.users.append(user)

    def remove_user(self, user: User):
        if user in self.users:
            self.users.remove(user)


@dataclass(**slots_params)
class RoomMessage:
    timestamp: int
    user: User
    message: str
    room: Room



from aiofiles import os as asyncos
import asyncio
from collections.abc import Callable
from concurrent.futures import Executor
from functools import partial
import logging
import mutagen
from mutagen.mp3 import BitrateMode
import os
import re
import sys
import time
from typing import Optional, Union
import uuid
from weakref import WeakSet

from ..base_manager import BaseManager
from .cache import SharesNullCache, SharesCache
from ..events import (
    EventBus,
    ScanCompleteEvent,
    SessionInitializedEvent,
    SessionDestroyedEvent,
    SharedDirectoryChangeEvent,
)
from ..exceptions import (
    FileNotFoundError,
    FileNotSharedError,
    SharedDirectoryError,
)
from .model import DirectoryShareMode, SharedDirectory, SharedItem
from ..naming import (
    chain_strategies,
    DefaultNamingStrategy,
    NumberDuplicateStrategy,
)
from ..network.network import Network
from ..protocol.messages import SharedFoldersFiles
from ..protocol.primitives import DirectoryData
from ..search.model import SearchQuery
from ..session import Session
from ..settings import Settings
from .utils import convert_items_to_file_data

logger = logging.getLogger(__name__)

ItemAttributes = list[tuple[int, int]]
ExecutorFactory = Callable[[], Executor]

_COMPRESSED_FORMATS = [
    'MP3',
    'MP4',
    'ASF',  # WMA
    'OggVorbis'
]
_LOSSLESS_FORMATS = [
    'FLAC',
    'WAVE'
]
_QUERY_CLEAN_PATTERN = re.compile(r"[\W_]")
"""Pattern to remove all non-word/digit characters from a string"""


def scan_directory(
        shared_directory: SharedDirectory,
        children: Optional[list[SharedDirectory]] = None) -> set[SharedItem]:
    """Scans the directory for items to share

    Warning: when using ProcessPoolExecutor on this method the returned items
    will not have the shared directory set in some cases. The shared_directory
    passed here is a copy and the object will be removed once this function
    finishes (see https://stackoverflow.com/a/72726998/1419478). In this case
    you should manually assign it again

    :param shared_directory: :class:`.SharedDirectory` instance
    :param children: list of :class:`.SharedDirectory` instances, the items in
        this list will not be returned and should be scanned individually
    :return: set of :class:`.SharedItem` objects found during the scan
    """
    children = children or []
    shared_items = set()
    for directory, _, files in os.walk(shared_directory.absolute_path):
        # Check if the currently scanned directory is part of any of the given
        # child directories
        abs_dir = os.path.normpath(os.path.abspath(directory))
        if any(child_dir.is_parent_of(abs_dir) for child_dir in children):
            continue

        subdir = os.path.relpath(directory, shared_directory.absolute_path)

        if subdir == '.':
            subdir = ''

        for filename in files:
            filepath = os.path.join(directory, filename)
            try:
                modified = os.path.getmtime(os.path.join(directory, filename))
            except OSError:
                logger.debug("could not get modified time for file : %r ", filepath)
            else:
                shared_items.add(
                    SharedItem(shared_directory, subdir, filename, modified)
                )

    return shared_items


def extract_attributes(filepath: str) -> ItemAttributes:
    """Attempts to extract attributes from the file at ``filepath``. If there was
    an error attempting to extract the attributes this method will log a warning
    and return an empty list of attributes
    """
    attributes = []
    try:
        mutagen_file = mutagen.File(filepath)
        if not mutagen_file:
            return []

        if mutagen_file.__class__.__name__ in _COMPRESSED_FORMATS:
            attributes = [
                (0, int(mutagen_file.info.bitrate / 1000), ),
                (1, int(mutagen_file.info.length), )
            ]

            # Figure out bitrate mode if applicable
            bitrate_mode = getattr(mutagen_file, 'bitrate_mode', BitrateMode.UNKNOWN)
            if bitrate_mode == BitrateMode.CBR:
                attributes.append((2, 0, ))
            elif bitrate_mode in (BitrateMode.VBR, BitrateMode.ABR, ):
                attributes.append((2, 1, ))

        elif mutagen_file.__class__.__name__ in _LOSSLESS_FORMATS:
            attributes = [
                (1, int(mutagen_file.info.length), ),
                (4, mutagen_file.info.sample_rate, ),
                (5, mutagen_file.info.bits_per_sample, ),
            ]

    except mutagen.MutagenError as exc:
        logger.warning(
            "failed retrieve audio file metadata. path=%r", filepath, exc_info=exc)

    return attributes


class SharesManager(BaseManager):
    _ALIAS_LENGTH = 5

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            network: Network, cache: Optional[SharesCache] = None,
            executor_factory: Optional[ExecutorFactory] = None):

        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._term_map: dict[str, WeakSet[SharedItem]] = {}
        self._shared_directories: list[SharedDirectory] = []
        self._session: Optional[Session] = None

        self.cache: SharesCache = cache if cache else SharesNullCache()
        self.executor: Optional[Executor] = None
        self.executor_factory: Optional[ExecutorFactory] = executor_factory

        self.naming_strategies = [
            DefaultNamingStrategy(),
            NumberDuplicateStrategy()
        ]

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)

    @property
    def shared_directories(self):
        return self._shared_directories

    def generate_alias(self, path: str, offset: int = 0) -> str:
        """Generates a directory alias for the given path, this method will be
        called recursively increasing the offset in case the alias is already
        found in the ``shared_directories``.

        The hardware address is mixed in to avoid getting the same alias for the
        same directory on different machines. Admittedly this is a lousy
        security measure but hopefully this will prevent easy leaking of files
        in case such issue would occur. Example: 'abcde' is always generated for
        'C:\\' so the attacker could guess where a file is located.

        :param path: the path for which to generate an alias
        :param offset: offset for the value of the initial 5 bytes (default=0)
        :return: a string of 5 alphabetic characters all lowercased
        """
        path_bytes = path.encode('utf8')
        unique_id = uuid.getnode().to_bytes(6, sys.byteorder)

        alias_bytearr = bytearray(
            [unique_id[-1] | offset for _ in range(self._ALIAS_LENGTH)])

        # Chunk the path into pieces of 5
        for c_idx in range(0, len(path_bytes), self._ALIAS_LENGTH):
            chunk = path_bytes[c_idx:c_idx + self._ALIAS_LENGTH]
            # previous_iter_byte XOR unique_id XOR current_iter_byte
            for b_idx, byte in enumerate(chunk):
                alias_bytearr[b_idx] ^= byte ^ unique_id[b_idx]

        # To lowercase
        alias_bytes = bytes([(byte % 26) + 0x61 for byte in alias_bytearr])
        alias_string = alias_bytes.decode('utf8')

        return alias_string

    async def start(self):
        if self.executor_factory:
            self.executor = self.executor_factory()

    async def stop(self) -> list[asyncio.Task]:
        if self.executor:
            self.executor.shutdown()
            self.executor = None

        return []

    async def load_data(self):
        self.read_cache()
        self.load_from_settings()

    async def store_data(self):
        self.write_cache()

    def load_from_settings(self):
        """Loads the shared directories from the settings. Existing directories
        will be updated, non-existing directories added, directories that no
        longer exist will be removed
        """
        new_shared_directories: list[SharedDirectory] = []

        # Add or update directories
        for directory_entry in self._settings.shares.directories:
            try:
                shared_directory = self.get_shared_directory(directory_entry.path)
            except SharedDirectoryError:
                shared_directory = self.add_shared_directory(
                    directory_entry.path,
                    share_mode=directory_entry.share_mode,
                    users=directory_entry.users
                )
            else:
                self.update_shared_directory(
                    shared_directory,
                    share_mode=directory_entry.share_mode,
                    users=directory_entry.users or []
                )

            new_shared_directories.append(shared_directory)

        self._shared_directories = new_shared_directories

        self.rebuild_term_map()

    def read_cache(self):
        """Read the directories from the cache"""
        logger.info("reading directories from cache")
        directories = self.cache.read()
        logger.info("read %d directories from cache", len(directories))
        self._shared_directories = directories

    def write_cache(self):
        """Write current shared directories to the cache"""
        directory_count = len(self._shared_directories)
        logger.info("writing %d directories to cache", directory_count)
        self.cache.write(self._shared_directories)
        logger.info("successfully wrote %d directories to cache", directory_count)

    def get_download_directory(self) -> str:
        """Gets the absolute path the to download directory configured from the
        settings

        :return: Absolute path of the value of the ``shares.download`` setting
        """
        download_dir = self._settings.shares.download
        return os.path.abspath(download_dir)

    async def create_directory(self, absolute_path: str):
        """Ensures the passed directory exists"""
        if not await asyncos.path.exists(absolute_path):
            logger.info("creating directory : %s", absolute_path)
            await asyncos.makedirs(absolute_path, exist_ok=True)

    def rebuild_term_map(self):
        logger.info("rebuilding term map")
        self._term_map = {}
        for shared_directory in self.shared_directories:
            self._build_term_map(shared_directory)

    def _build_term_map(self, shared_directory: SharedDirectory):
        """Builds a list of valid terms for the given shared directory"""
        for item in shared_directory.items:
            self._add_item_to_term_map(item)
        logger.debug("term map contains %d terms", len(self._term_map))

    def get_shared_item_cache(self, remote_path: str, username: Optional[str] = None) -> SharedItem:
        """Gets a shared item from the cache if it exists

        If a ``username`` is passed this will also check if the file is locked
        and raise a :class:`.FileNotSharedError` if the file is not accessible
        for that user

        :param remote_path: the remote_path
        :param username: optional username to check if the file is shared or not
        :raise FileNotFoundError: filename was not found in shared_items
        :raise FileNotSharedError: file is found, but locked for the given
            ``username``
        """
        for shared_directory in self._shared_directories:
            try:
                item = shared_directory.get_item_by_remote_path(remote_path)
            except FileNotFoundError:
                pass
            else:
                if username and self.is_item_locked(item, username):
                    raise FileNotSharedError(f"file is not shared to user {username}")
                return item
        else:
            raise FileNotFoundError(f"file name {remote_path} not found in cache")

    async def get_shared_item(self, remote_path: str, username: Optional[str] = None) -> SharedItem:
        """Same as :meth:`get_shared_item_cache` but also checks if the file
        exists on disk

        :param remote_path: the remote_path
        :param username: optional username to check if the file is shared or not
        :raise FileNotFoundError: filename was not found in shared_items or was
            found but did not exist on disk
        :raise FileNotSharedError: file is found, but locked for the given
            ``username``
        """
        item = self.get_shared_item_cache(remote_path, username=username)
        if not await asyncos.path.exists(item.get_absolute_path()):
            raise FileNotFoundError(
                f"file with remote_path {remote_path} found in cache but not on disk"
            )
        return item

    def find_shared_item_cache(
            self, remote_path: str, username: Optional[str] = None) -> Optional[SharedItem]:
        """Equivelant to :meth:`get_shared_item_cache` but returns ``None`` if the
        shared item is not found
        """
        try:
            return self.get_shared_item_cache(remote_path, username=username)
        except (FileNotFoundError, FileNotSharedError):
            return None

    async def find_shared_item(
            self, remote_path: str, username: Optional[str] = None) -> Optional[SharedItem]:
        """Equivelant to :meth:`get_shared_item` but returns ``None`` if the
        shared item is not found
        """
        try:
            return await self.get_shared_item(remote_path, username=username)
        except (FileNotFoundError, FileNotSharedError):
            return None

    def add_shared_directory(
            self, shared_directory: str,
            share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE,
            users: Optional[list[str]] = None) -> SharedDirectory:
        """Adds a shared directory. This method will call :meth:`generate_alias`
        and add the directory to the directory map.

        This method will not scan the directory, however if the directory is a
        (sub)child of an already registered shared directory the items from that
        directory will be removed from the parent directory to the new directory

        For scanning see the :meth:`scan`, :meth:`scan_directory_files` and
        :meth:`scan_directory_file_attributes` methods.

        :param shared_directory: path of the shared directory
        :param share_mode: the share mode for the directory
        :param users: in case the share mode is `USERS`, a list of users to
            share it with
        :return: a :class:`.SharedDirectory` object
        :raise SharedDirectoryError: if the directory is already shared
        """
        if self.is_directory_shared(shared_directory):
            raise SharedDirectoryError(
                f"directory {shared_directory} is already shared")

        # Generate an absolute path, alias and create the object
        abs_directory = os.path.normpath(os.path.abspath(shared_directory))
        alias = self.generate_alias(abs_directory)

        directory_object = SharedDirectory(
            shared_directory,
            abs_directory,
            alias,
            share_mode=share_mode,
            users=users or []
        )

        # If the new directory is a child of an existing directory, move items
        # to the child directory and remove them from the parent directory
        parents = self._get_parent_directories(directory_object)
        if parents:
            parent = parents[-1]
            children = parent.get_items_for_directory(directory_object)
            directory_object.items |= children
            parent.items -= children

        self._shared_directories.append(directory_object)

        self._event_bus.emit_sync(SharedDirectoryChangeEvent(directory_object))

        return directory_object

    def update_shared_directory(
            self, directory: Union[str, SharedDirectory],
            share_mode: Optional[DirectoryShareMode] = None,
            users: Optional[list[str]] = None) -> SharedDirectory:
        """Updates ``share_mode`` and ``users`` values for the given directory

        :param directory: if a string is given this method will attempt to find
            the shared directory based on the absolute path
        :return: the updated directory
        :raise SharedDirectoryError: if the given directory is a string and not
            in the list of shared directories
        """
        if isinstance(directory, str):
            shared_directory = self.get_shared_directory(directory)
        else:
            shared_directory = directory

        if share_mode is not None:
            shared_directory.share_mode = share_mode

        if users is not None:
            shared_directory.users = users

        self._event_bus.emit_sync(SharedDirectoryChangeEvent(shared_directory))

        return shared_directory

    def remove_shared_directory(self, directory: Union[str, SharedDirectory]) -> SharedDirectory:
        """Removes the given shared directory. If the directory was a
        subdirectory of another shared directory its items will be moved into
        that directory

        Example file structure:

        * Music\\

            * Artist_One\\

                * song_one.mp3

            * Artist_Two\\

                * song_two.mp3

        If the user shares 2 directories:

        * Music\\ : EVERYONE
        * Music\\Artist_One\\ : FRIENDS

        And removes the `Music\\Artist_One\\ shared directory the files of that
        directory get returned back to the parent directory. In this case file
        ``song_two.mp3`` will be shared with EVERYONE again

        :param shared_directory: :class:`.SharedDirectory` instance to remove
        :return: the removed directory
        :raise SharedDirectoryError: raised when the passed `directory` was not
            added to the manager
        """
        if isinstance(directory, str):
            shared_directory = self.get_shared_directory(directory)
        else:
            if directory not in self._shared_directories:
                raise SharedDirectoryError(
                    "attempted to remove directory which was not added to the manager"
                )
            else:
                shared_directory = directory

        self._shared_directories.remove(shared_directory)

        parents = self._get_parent_directories(shared_directory)
        # If the directory has a parent directory, move all items into that
        # directory
        if parents:
            parent = parents[-1]
            parent.items |= shared_directory.items

        self._cleanup_term_map()

        self._event_bus.emit_sync(SharedDirectoryChangeEvent(shared_directory))

        return shared_directory

    def get_shared_directory(self, directory: str) -> SharedDirectory:
        """Calculates the absolute path of given ``directory`` and looks for the
        matching :class:`.SharedDirectory` instance

        :raise SharedDirectoryError: if there is no corresponding
            :class:`.SharedDirectory` instance registered in this class
        """
        abs_path = os.path.normpath(os.path.abspath(directory))
        for shared_directory in self.shared_directories:
            if shared_directory.absolute_path == abs_path:
                return shared_directory

        raise SharedDirectoryError(
            f"did not find shared directory with path : {directory}")

    def is_directory_shared(self, directory: str) -> bool:
        """Checks if a directory is already a shared directory by checking the
        absolute path of that directory
        """
        try:
            self.get_shared_directory(directory)
        except SharedDirectoryError:
            return False
        else:
            return True

    async def scan_directory_files(self, shared_directory: SharedDirectory):
        """Scans the files for the given ``shared_directory``

        :param shared_directory: :class:`.SharedDirectory` instance to scan
        :raise SharedDirectoryError: raised when the passed ``shared_directory``
            was not added to the manager
        """
        loop = asyncio.get_running_loop()

        if shared_directory not in self._shared_directories:
            raise SharedDirectoryError(
                "attempted to scan directory which was not added to the manager"
            )

        logger.info("scheduling scan for directory : %r", shared_directory)
        try:
            shared_items: set[SharedItem] = await loop.run_in_executor(
                self.executor,
                partial(
                    scan_directory,
                    shared_directory,
                    children=self._get_child_directories(shared_directory)
                )
            )
        except Exception:
            logger.exception("exception scanning directory : %r", shared_directory)

        else:
            logger.debug("scan found %d files for directory : %r", len(shared_items), shared_directory)

            # When using a ProcessPoolExecutor the `shared_directory` property
            # might be set to None (since objects are cloned instead of passed
            # by reference). Re-assign it the proper object
            for item in shared_items:
                item.shared_directory = shared_directory

            # Adds all new items to the directory items
            shared_directory.items |= shared_items

            # Remove all items from the cache that weren't in the returned items
            # set. Because the `modified` parameter is part of the hash items
            # with a changed `modified` parameter will also be removed meaning
            # their attributes will be reset and these files attributes need
            # to be rescanned
            shared_directory.items -= (shared_directory.items ^ shared_items)

        self._build_term_map(shared_directory)
        self._cleanup_term_map()

    async def scan_directory_file_attributes(self, shared_directory: SharedDirectory):
        """Scans the file attributes for files in the given ``shared_directory``
        only files that do not have attributes will be scanned

        The results of the scan are handled internally and are automatically to
        the :class:`.SharedItem` object for which the scan was performed

        :param shared_directory: :class:`.SharedDirectory` instance for which
            the files need to be scanned
        """
        loop = asyncio.get_running_loop()

        # Schedule the items on the executor
        futures: list[asyncio.Future] = []
        for item in shared_directory.items:
            if item.attributes is None:
                future = loop.run_in_executor(
                    self.executor,
                    partial(extract_attributes, item.get_absolute_path())
                )
                future.add_done_callback(
                    partial(self._extract_attributes_callback, item)
                )
                futures.append(future)

        logger.debug(
            "scheduled %d / %d items for attribute extracting for directory %s",
            len(futures), len(shared_directory.items), shared_directory
        )

        start = time.perf_counter()

        await asyncio.gather(*futures, return_exceptions=True)

        logger.debug(
            "scanned attributes for %d items in %f s", len(futures), time.perf_counter() - start)

    def _extract_attributes_callback(self, item: SharedItem, future: asyncio.Future):
        try:
            item.attributes = future.result()
        except Exception:
            logger.warning("exception fetching shared item attributes")

    async def scan(self):
        """Scan the files and their attributes for all directories currently
        defined in the ``shared_directories``

        This method will emit a :class:`.ScanCompleteEvent` on the event bus
        and report the shares to the server
        """
        start_time = time.perf_counter()

        files_tasks = [
            self.scan_directory_files(shared_directory)
            for shared_directory in self._shared_directories
        ]
        await asyncio.gather(*files_tasks, return_exceptions=True)

        attribute_futures = [
            self.scan_directory_file_attributes(shared_directory)
            for shared_directory in self._shared_directories
        ]
        await asyncio.gather(*attribute_futures)

        logger.info("completed scan in %f seconds", time.perf_counter() - start_time)
        folder_count, file_count = self.get_stats()
        await self._event_bus.emit(
            ScanCompleteEvent(folder_count, file_count)
        )

        await self.report_shares()

    async def get_filesize(self, shared_item: SharedItem) -> int:
        return await asyncos.path.getsize(shared_item.get_absolute_path())

    def query(
            self, query: Union[str, SearchQuery],
            username: Optional[str] = None,
            excluded_search_phrases: Optional[list[str]] = None) -> tuple[list[SharedItem], list[SharedItem]]:
        """Performs a query on the ``shared_directories`` returning the matching
        items. If ``username`` is passed this method will return a list of
        visible results and list of locked results. If ``None`` the second list
        will always be empty.

        This method makes a first pass using the built in term map and filters
        the remaining results using regular expressions.

        :param query: the query to perform on the shared directories
        :param username: optionally the username of the user making the query.
            This is used to determine locked results, if not given the locked
            results list will be empty
        :param excluded_search_phrases: optional list of search phrases that
            should be excluded from the search results
        :return: two lists of visible results and locked results
        """
        if not isinstance(query, SearchQuery):
            search_query = SearchQuery.parse(query)
        else:
            search_query = query

        # Ignore if no valid include or wildcard terms are given
        if not search_query.has_inclusion_terms():
            return [], []

        # First round using the term map
        include_terms = []
        for term in search_query.include_terms:
            subterms = re.split(_QUERY_CLEAN_PATTERN, term)
            for subterm in subterms:
                if not subterm:
                    continue

                if subterm not in self._term_map:  # Optimization
                    return [], []

                include_terms.append(subterm)

        for term in search_query.wildcard_terms:
            subterms = re.split(_QUERY_CLEAN_PATTERN, term)
            for idx, subterm in enumerate(subterms):
                if not subterm:
                    continue

                if idx == 0:
                    matching_terms = [
                        map_term for map_term in self._term_map.keys()
                        if map_term.endswith(subterm)
                    ]

                    if not matching_terms:  # Optimization
                        return [], []

                    include_terms.extend(matching_terms)
                else:
                    if subterm not in self._term_map:  # Optimization
                        return [], []

                    include_terms.append(subterm)

        found_items = set(self._term_map[include_terms[0]])
        for include_term in include_terms:
            found_items &= set(self._term_map[include_term])

        # Regular expressions on the remaining items

        to_keep = set()
        excl_phrases = excluded_search_phrases or []
        for found_item in found_items:

            if not all(matcher(found_item.get_query_path()) for matcher in search_query.matchers_iter()):
                continue

            # Excluded search phrases
            for excl_phrase in excl_phrases:
                if excl_phrase in found_item.get_query_path().lower():
                    logger.debug(
                        "removing search result %r due to excluded phrase %r",
                        found_item.get_absolute_path(), excl_phrase
                    )
                    break

            else:
                to_keep.add(found_item)

            if len(to_keep) >= self._settings.searches.receive.max_results:
                break

        found_items = to_keep

        # Order by visible and locked results when username is given
        if username:
            visible_results = []
            locked_results = []
            for item in found_items:
                if self.is_item_locked(item, username):
                    locked_results.append(item)
                else:
                    visible_results.append(item)
            return visible_results, locked_results
        else:
            return list(found_items), []

    def get_stats(self) -> tuple[int, int]:
        """Gets the total amount of shared directories and files.

        :return: directory and file count as a ``tuple``
        """
        file_count = sum(
            len(directory.items) for directory in self._shared_directories
        )
        dir_count = sum(
            len(set(item.subdir for item in directory.items))
            for directory in self._shared_directories
        )
        return dir_count, file_count

    def calculate_download_path(self, remote_path: str) -> tuple[str, str]:
        """Calculates the local download path for a remote path returned by
        another peer.

        :return: ``tuple`` of the directory and file name
        """
        download_dir = self.get_download_directory()

        return chain_strategies(
            self.naming_strategies,
            remote_path,
            download_dir
        )

    def get_shared_directories_for_user(self, username: str) -> tuple[list[SharedDirectory], list[SharedDirectory]]:
        public_dirs = []
        locked_dirs = []
        for shared_dir in self._shared_directories:
            if self.is_directory_locked(shared_dir, username):
                locked_dirs.append(shared_dir)
            else:
                public_dirs.append(shared_dir)

        return public_dirs, locked_dirs

    def create_shares_reply(self, username: str) -> tuple[list[DirectoryData], list[DirectoryData]]:
        """Creates a complete list of the currently shared items as a reply to
        a :class:`.PeerSharesRequest` messages

        :param username: username of the user requesting the shares reply, this
            is used to determine the locked results
        :return: ``tuple`` with two lists: public directories and locked
            directories
        """
        def list_unique_directories(directories: list[SharedDirectory]) -> dict[tuple[str, ...], list[SharedItem]]:
            response_dirs: dict[tuple[str, ...], list[SharedItem]] = {}

            for directory in directories:
                for item in directory.items:
                    dir_path = item.get_remote_directory_path_parts()

                    # Create all possible subdirectory paths
                    for x in range(len(dir_path)):
                        dir_path_subpart = dir_path[:x + 1]
                        if dir_path_subpart not in response_dirs:
                            response_dirs[dir_path_subpart] = []

                    response_dirs[dir_path].append(item)

            return response_dirs

        def convert_to_directory_shares(directory_map: dict[tuple[str, ...], list[SharedItem]]) -> list[DirectoryData]:
            public_shares = []
            for directory, files in directory_map.items():
                public_shares.append(
                    DirectoryData(
                        name='\\'.join(directory),
                        files=convert_items_to_file_data(files, use_full_path=False)
                    )
                )
            return public_shares

        visible_dirs, locked_dirs = self.get_shared_directories_for_user(username)
        visible_shares = convert_to_directory_shares(
            list_unique_directories(visible_dirs))
        locked_shares = convert_to_directory_shares(
            list_unique_directories(locked_dirs))

        return visible_shares, locked_shares

    def create_directory_reply(self, remote_directory: str) -> list[DirectoryData]:
        """Lists directory data as a response to a directory request. This will
        not contain any information about subdirectories, only the files within
        that directory

        This method differs from the official clients who don't respond at all
        to the message if the directory does not exist. Instead if the directory
        does not exist and empty list is returned

        :param remote_directory: remote path of the directory
        :return: list of directories. Empty if the directory is not shared, a
            list with one entry if the directory is found
        """
        items: list[SharedItem] = []

        remote_dir_parts = tuple(remote_directory.split('\\'))
        remote_dir_parts_len = len(remote_dir_parts)
        is_shared = False

        for shared_dir in self._shared_directories:
            for item in shared_dir.items:
                # Funny logic to determine if the directory is shared
                if not is_shared:
                    if item.get_remote_directory_path_parts()[:remote_dir_parts_len] == remote_dir_parts:
                        is_shared = True

                if item.get_remote_directory_path() == remote_directory:
                    items.append(item)

        if not is_shared:
            return []

        return [
            DirectoryData(
                name=remote_directory,
                files=convert_items_to_file_data(items, use_full_path=False)
            )
        ]

    def _add_item_to_term_map(self, item: SharedItem):
        path = (item.subdir + "/" + item.filename).lower()
        terms = re.split(_QUERY_CLEAN_PATTERN, path)
        for term in terms:
            if not term:
                continue

            if term not in self._term_map:
                self._term_map[term] = WeakSet()
            self._term_map[term].add(item)

    def _cleanup_term_map(self):
        self._term_map = {
            term: values for term, values in self._term_map.items()
            if len(values) > 0
        }

    def _get_parent_directories(self, shared_directory: SharedDirectory) -> list[SharedDirectory]:
        """Returns a list of parent shared directories. The parent directories
        will be sorted by length of the absolute path (longest last)
        """
        parent_dirs = [
            directory for directory in self._shared_directories
            if directory != shared_directory and directory.is_parent_of(shared_directory)
        ]
        return sorted(parent_dirs, key=lambda d: len(d.absolute_path))

    def _get_child_directories(self, shared_directory: SharedDirectory):
        """Returns a list of child directories"""
        children = [
            directory for directory in self._shared_directories
            if directory != shared_directory and directory.is_child_of(shared_directory)
        ]
        return children

    def is_directory_locked(self, directory: SharedDirectory, username: str) -> bool:
        """Checks if the shared directory is locked for the given ``username``
        """
        if directory.share_mode == DirectoryShareMode.FRIENDS:
            return username not in self._settings.users.friends
        elif directory.share_mode == DirectoryShareMode.USERS:
            return username not in directory.users
        return False

    def is_item_locked(self, item: SharedItem, username: str) -> bool:
        """Checks if the shared item is locked for the given ``username``"""
        return self.is_directory_locked(item.shared_directory, username)

    async def report_shares(self):
        """Reports the shares amount to the server"""
        if not self._session:
            return

        folder_count, file_count = self.get_stats()
        logger.debug("reporting shares (folder_count=%d, file_count=%d)", folder_count, file_count)
        await self._network.send_server_messages(
            SharedFoldersFiles.Request(
                shared_folder_count=folder_count,
                shared_file_count=file_count
            )
        )

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session
        await self.report_shares()

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None



from collections.abc import Set
import os
from typing import Annotated, Any, Optional, Sequence

from pydantic import BaseModel, BeforeValidator, Field, model_validator
from pydantic_settings import BaseSettings

from .constants import (
    UPNP_DEFAULT_CHECK_INTERVAL,
    UPNP_DEFAULT_LEASE_DURATION,
    UPNP_DEFAULT_SEARCH_TIMEOUT,
)
from .network.network import ListeningConnectionErrorMode, PeerConnectMode
from .shares.model import DirectoryShareMode
from .user.model import BlockingFlag


def translate_blocked_users(value: Any) -> dict[str, BlockingFlag]:
    """Converts blocked user set to a dictionary object"""
    if isinstance(value, dict):
        return value

    if isinstance(value, (Sequence, Set)):
        blocked_dict = {}
        for username in value:
            blocked_dict[username] = BlockingFlag.ALL

        return blocked_dict

    raise ValueError(f"unable to translate type {type(value)}")


class SharedDirectorySettingEntry(BaseModel, validate_assignment=True):
    path: str
    share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE
    users: list[str] = Field(default_factory=list)


class WishlistSettingEntry(BaseModel, validate_assignment=True):
    query: str
    enabled: bool = True


class UpnpSettings(BaseModel, validate_assignment=True):
    enabled: bool = True
    lease_duration: int = UPNP_DEFAULT_LEASE_DURATION
    check_interval: int = UPNP_DEFAULT_CHECK_INTERVAL
    search_timeout: int = UPNP_DEFAULT_SEARCH_TIMEOUT


class ReconnectSettings(BaseModel, validate_assignment=True):
    auto: bool = False
    timeout: int = 10


class ServerSettings(BaseModel, validate_assignment=True):
    hostname: str = 'server.slsknet.org'
    port: int = 2416
    reconnect: ReconnectSettings = ReconnectSettings()


class ListeningSettings(BaseModel, validate_assignment=True):
    error_mode: ListeningConnectionErrorMode = ListeningConnectionErrorMode.CLEAR
    port: int = 60000
    obfuscated_port: int = 60001


class PeerSettings(BaseModel, validate_assignment=True):
    obfuscate: bool = False
    connect_mode: PeerConnectMode = PeerConnectMode.RACE


class NetworkLimitSettings(BaseModel, validate_assignment=True):
    upload_speed_kbps: int = 0
    download_speed_kbps: int = 0


class NetworkSettings(BaseModel, validate_assignment=True):
    server: ServerSettings = Field(default_factory=ServerSettings)
    listening: ListeningSettings = Field(default_factory=ListeningSettings)
    peer: PeerSettings = Field(default_factory=PeerSettings)
    upnp: UpnpSettings = Field(default_factory=UpnpSettings)
    limits: NetworkLimitSettings = Field(default_factory=NetworkLimitSettings)


class UserInfoSettings(BaseModel, validate_assignment=True):
    description: Optional[str] = None
    picture: Optional[bytes] = None


class CredentialsSettings(BaseModel, validate_assignment=True):
    username: str
    password: str
    info: UserInfoSettings = Field(default_factory=UserInfoSettings)

    def are_configured(self) -> bool:
        """Returns whether the credentials are correctly configured"""
        return bool(self.username) and self.password is not None


class SearchSendSettings(BaseModel, validate_assignment=True):
    store_results: bool = True
    request_timeout: int = 0
    wishlist_request_timeout: int = -1


class SearchReceiveSettings(BaseModel, validate_assignment=True):
    store_amount: int = 500
    max_results: int = 100


class SearchSettings(BaseModel, validate_assignment=True):
    send: SearchSendSettings = Field(default_factory=SearchSendSettings)
    receive: SearchReceiveSettings = Field(default_factory=SearchReceiveSettings)
    wishlist: list[WishlistSettingEntry] = Field(default_factory=list)

    @model_validator(mode='before')
    @classmethod
    def handle_max_result_alias(cls, values: dict):
        if 'max_results' in values:
            value = values.pop('max_results')
            if 'receive' not in values:
                values['receive'] = {}

            values['receive']['max_results'] = value

        return values

    @property
    def max_results(self) -> int:
        return self.receive.max_results

    @max_results.setter
    def max_results(self, value: int):
        self.receive.max_results = value


class TransferLimitSettings(BaseModel, validate_assignment=True):
    upload_slots: int = 2


class TransfersSettings(BaseModel, validate_assignment=True):
    limits: TransferLimitSettings = Field(default_factory=TransferLimitSettings)
    report_interval: float = 0.250


class SharesSettings(BaseModel, validate_assignment=True):
    scan_on_start: bool = True
    download: str = os.getcwd()
    directories: list[SharedDirectorySettingEntry] = Field(default_factory=list)


class RoomsSettings(BaseModel, validate_assignment=True):
    auto_join: bool = True
    private_room_invites: bool = True
    favorites: set[str] = Field(default_factory=set)


class UsersSettings(BaseModel, validate_assignment=True):
    friends: set[str] = Field(default_factory=set)
    blocked: Annotated[
        dict[str, BlockingFlag],
        BeforeValidator(translate_blocked_users)
    ] = Field(default_factory=dict)

    def is_blocked(self, username: str, flag: BlockingFlag) -> bool:
        return bool(self.blocked.get(username, BlockingFlag.NONE) & flag)


class InterestsSettings(BaseModel, validate_assignment=True):
    liked: set[str] = Field(default_factory=set)
    hated: set[str] = Field(default_factory=set)


class DebugSettings(BaseModel, validate_assignment=True):
    search_for_parent: bool = True
    ip_overrides: dict[str, str] = Field(default_factory=dict)
    log_connection_count: bool = False


class Settings(BaseSettings, validate_assignment=True):
    network: NetworkSettings = Field(default_factory=NetworkSettings)
    credentials: CredentialsSettings = Field(default_factory=CredentialsSettings)  # type: ignore
    searches: SearchSettings = Field(default_factory=SearchSettings)
    shares: SharesSettings = Field(default_factory=SharesSettings)
    users: UsersSettings = Field(default_factory=UsersSettings)
    rooms: RoomsSettings = Field(default_factory=RoomsSettings)
    interests: InterestsSettings = Field(default_factory=InterestsSettings)
    transfers: TransfersSettings = Field(default_factory=TransfersSettings)
    debug: DebugSettings = Field(default_factory=DebugSettings)


def _main(args):
    import json
    import sys

    if args.command == 'generate':
        settings = Settings(
            credentials=CredentialsSettings(
                username=args.username,
                password=args.password
            )
        )

        dump = settings.model_dump(mode='json')
        json.dump(dump, sys.stdout, indent=2)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Setting file tools")
    subparsers = parser.add_subparsers(dest='command')

    gen_parser = subparsers.add_parser('generate')
    gen_parser.add_argument('--username', '-u', default="TestUser")
    gen_parser.add_argument('--password', '-p', default="Pass0")

    args = parser.parse_args()
    _main(args)



import logging

from .base_manager import BaseManager
from .events import (
    on_message,
    build_message_map,
    EventBus,
    MessageReceivedEvent,
    UserDirectoryEvent,
    UserSharesReplyEvent,
)
from .protocol.messages import (
    PeerDirectoryContentsRequest,
    PeerDirectoryContentsReply,
    PeerSharesRequest,
    PeerSharesReply,
    PeerUserInfoReply,
    PeerUserInfoRequest,
)
from .network.connection import PeerConnection
from .network.network import Network
from .settings import Settings
from .shares.manager import SharesManager
from .transfer.interface import UploadInfoProvider
from .user.manager import UserManager
from .user.model import BlockingFlag
from .utils import ticket_generator


logger = logging.getLogger(__name__)


class PeerManager(BaseManager):
    """Peer manager is responsible for handling peer messages"""

    def __init__(
            self, settings: Settings,
            event_bus: EventBus,
            user_manager: UserManager,
            shares_manager: SharesManager, upload_info_provider: UploadInfoProvider,
            network: Network):

        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._user_manager: UserManager = user_manager
        self._shares_manager: SharesManager = shares_manager
        self._upload_info_provider: UploadInfoProvider = upload_info_provider

        self._ticket_generator = ticket_generator()

        self._MESSAGE_MAP = build_message_map(self)

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)

    # Peer messages

    @on_message(PeerSharesRequest.Request)
    async def _on_peer_shares_request(self, message: PeerSharesRequest.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        if self._settings.users.is_blocked(connection.username, BlockingFlag.SHARES):
            return

        visible, locked = self._shares_manager.create_shares_reply(connection.username)
        await connection.send_message(
            PeerSharesReply.Request(
                directories=visible,
                locked_directories=locked
            )
        )

    @on_message(PeerSharesReply.Request)
    async def _on_peer_shares_reply(self, message: PeerSharesReply.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        logger.info(
            "PeerSharesReply : from username %s, got %d directories",
            connection.username, len(message.directories)
        )

        user = self._user_manager.get_user_object(connection.username)

        await self._event_bus.emit(
            UserSharesReplyEvent(
                user=user,
                directories=message.directories,
                locked_directories=message.locked_directories or [],
                raw_message=message
            )
        )

    @on_message(PeerDirectoryContentsRequest.Request)
    async def _on_peer_directory_contents_req(
            self, message: PeerDirectoryContentsRequest.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerDirectoryContentsRequest for a connection that wasn't properly initialized")
            return

        if self._settings.users.is_blocked(connection.username, BlockingFlag.SHARES):
            return

        directories = self._shares_manager.create_directory_reply(message.directory)
        await connection.send_message(
            PeerDirectoryContentsReply.Request(
                ticket=message.ticket,
                directory=message.directory,
                directories=directories
            )
        )

    @on_message(PeerDirectoryContentsReply.Request)
    async def _on_peer_directory_contents_reply(
            self, message: PeerDirectoryContentsReply.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerDirectoryContentsReply for a connection that wasn't properly initialized")
            return

        user = self._user_manager.get_user_object(connection.username)
        await self._event_bus.emit(
            UserDirectoryEvent(
                user=user,
                directory=message.directory,
                directories=message.directories,
                raw_message=message
            )
        )

    @on_message(PeerUserInfoRequest.Request)
    async def _on_peer_user_info_request(self, message: PeerUserInfoRequest.Request, connection: PeerConnection):

        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        if self._settings.users.is_blocked(connection.username, BlockingFlag.INFO):
            return

        description = self._settings.credentials.info.description or ""
        picture = self._settings.credentials.info.picture

        await connection.send_message(
            PeerUserInfoReply.Request(
                description=description,
                has_picture=bool(picture),
                picture=picture,
                upload_slots=self._upload_info_provider.get_upload_slots(),
                queue_size=self._upload_info_provider.get_queue_size(),
                has_slots_free=self._upload_info_provider.has_slots_free()
            )
        )

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)



from collections.abc import Callable, Generator
from dataclasses import dataclass, field
import datetime
from enum import auto, Enum
import logging
import re
from typing import Optional

from ..protocol.primitives import FileData
from ..shares.utils import create_term_pattern
from ..tasks import Timer


logger = logging.getLogger(__name__)


class SearchType(Enum):
    NETWORK = auto()
    USER = auto()
    ROOM = auto()
    WISHLIST = auto()


@dataclass(slots=True)
class ReceivedSearch:
    """Used for keeping track of searches received from the distributed parent
    or server
    """
    username: str
    query: str
    result_count: int


@dataclass(slots=True)
class SearchQuery:
    query: str
    include_terms: set[str] = field(default_factory=set)
    exclude_terms: set[str] = field(default_factory=set)
    wildcard_terms: set[str] = field(default_factory=set)

    @classmethod
    def parse(cls, query: str) -> 'SearchQuery':
        """Parses the given query string into terms and creates a new
        :class:`.SearchQuery` object
        """
        obj = cls(query)

        terms = query.split()
        for term in terms:
            # Ignore terms containing only non-word chars
            l_term = term.lower()
            if not re.search(r'[^\W_]', l_term):
                continue

            if term.startswith('*'):
                obj.wildcard_terms.add(l_term[1:])
            elif term.startswith('-'):
                obj.exclude_terms.add(l_term[1:])
            else:
                obj.include_terms.add(l_term)

        return obj

    def matchers_iter(self) -> Generator[Callable[[str], bool], None, None]:
        """Generator for term matchers"""
        for include_term in self.include_terms:
            pattern = create_term_pattern(include_term, wildcard=False)
            yield lambda fn: bool(pattern.search(fn))

        for wildcard_term in self.wildcard_terms:
            pattern = create_term_pattern(wildcard_term, wildcard=True)
            yield lambda fn: bool(pattern.search(fn))

        for exclude_term in self.exclude_terms:
            pattern = create_term_pattern(exclude_term, wildcard=False)
            yield lambda fn: not pattern.search(fn)

    def has_inclusion_terms(self) -> bool:
        """Return whether this query has any valid inclusion terms"""
        return bool(self.include_terms) or bool(self.wildcard_terms)


@dataclass(slots=True)
class SearchResult:
    """Search result received from a user"""
    ticket: int
    username: str

    has_free_slots: bool = False
    avg_speed: int = 0
    queue_size: int = 0

    shared_items: list[FileData] = field(default_factory=list)
    locked_results: list[FileData] = field(default_factory=list)


@dataclass(slots=True)
class SearchRequest:
    """Search request we have made"""
    ticket: int
    query: str
    search_type: SearchType = SearchType.NETWORK

    room: Optional[str] = None
    username: Optional[str] = None
    results: list[SearchResult] = field(default_factory=list)
    started: datetime.datetime = field(default_factory=datetime.datetime.now)

    timer: Optional[Timer] = None



import os
import shelve
from typing import Protocol
from .model import SharedDirectory


class SharesCache(Protocol):
    """Abstract base class for storing shares"""

    def read(self) -> list[SharedDirectory]:
        ...

    def write(self, shared_directories: list[SharedDirectory]):
        ...


class SharesNullCache:

    def read(self) -> list[SharedDirectory]:  # pragma: no cover
        return []

    def write(self, shared_directories: list[SharedDirectory]):  # pragma: no cover
        pass


class SharesShelveCache:
    """Shares cache that uses the Python built-in `shelve` module"""
    DEFAULT_FILENAME = 'shares_index'

    def __init__(self, data_directory: str, filename: str = DEFAULT_FILENAME):
        self.data_directory: str = data_directory
        self.filename: str = filename

    def _get_index_path(self) -> str:
        return os.path.join(self.data_directory, self.filename)

    def read(self) -> list[SharedDirectory]:
        with shelve.open(self._get_index_path(), 'c') as db:
            directories: list[SharedDirectory] = db.get('index', list())
            for directory in directories:
                new_items = set()
                for item in directory.items:
                    item.shared_directory = directory
                    new_items.add(item)
                directory.items = new_items
            return directories

    def write(self, shared_directories: list[SharedDirectory]):
        with shelve.open(self._get_index_path(), 'c') as db:
            db['index'] = shared_directories
